/*
 * XML libary for MySQL.
 * A set of MySQL user defined functions (udf) to create XML output directly from MySQL using an SQL query
 *
 * Copyright (C) 2006 Arnold Daniels <info@adaniels.nl>
 * web:   http://www.mysqludf.com/lib_mysqludf_xql
 * email: mysql-udf-repository@googlegroups.com
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#define LIBVERSION "lib_mysqludf_xql version 0.9.4"

#if defined(_WIN32) || defined(_WIN64) || defined(__WIN32__) || defined(WIN32)
#define DLLEXP __declspec(dllexport) 
#else
#define DLLEXP
#endif

#ifdef STANDARD
/* STANDARD is defined, don't use any mysql functions */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef __WIN__
typedef unsigned __int64 ulonglong;	/* Microsofts 64 bit types */
typedef __int64 longlong;
#else
typedef unsigned long long ulonglong;
typedef long long longlong;
#endif /*__WIN__*/
#else
#include <my_global.h>
#include <my_sys.h>
#include <m_string.h>		// To get strcpy()
#endif
#include <mysql.h>
#include <ctype.h>

#include <libxml/tree.h>
#include <libxml/xmlwriter.h>
#include <libxml/uri.h>

#ifdef HAVE_DLOPEN

/* These must be right or mysqld will not find the symbol! */

#ifdef	__cplusplus
extern "C" {
#endif
	DLLEXP my_bool lib_mysqludf_xql_info_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
	DLLEXP void lib_mysqludf_xql_info_deinit(UDF_INIT *initid);
	DLLEXP char *lib_mysqludf_xql_info(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);

	DLLEXP my_bool xql_test_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
	DLLEXP void xql_test_deinit(UDF_INIT *initid);
	DLLEXP char *xql_test(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);

	DLLEXP my_bool xql_textdecl_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
	DLLEXP void xql_textdecl_deinit(UDF_INIT *initid);
	DLLEXP char *xql_textdecl(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);

	DLLEXP my_bool xql_pi_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
	DLLEXP void xql_pi_deinit(UDF_INIT *initid);
	DLLEXP char *xql_pi(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);

	DLLEXP my_bool xql_dtd_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
	DLLEXP void xql_dtd_deinit(UDF_INIT *initid);
	DLLEXP char *xql_dtd(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);


	DLLEXP my_bool xql_element_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
	DLLEXP void xql_element_deinit(UDF_INIT *initid);
	DLLEXP char *xql_element(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);

	DLLEXP my_bool xql_forest_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
	DLLEXP void xql_forest_deinit(UDF_INIT *initid);
	DLLEXP char *xql_forest(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);

	DLLEXP my_bool xql_concat_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
	DLLEXP void xql_concat_deinit(UDF_INIT *initid);
	DLLEXP char *xql_concat(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);

	DLLEXP my_bool xql_agg_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
	DLLEXP void xql_agg_deinit(UDF_INIT *initid);
	DLLEXP void xql_agg_reset(UDF_INIT* initid, UDF_ARGS* args, char* is_null, char* message);
	DLLEXP void xql_agg_clear(UDF_INIT* initid, char* is_null, char* message);
	DLLEXP void xql_agg_add(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error);
	DLLEXP char *xql_agg(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);


	DLLEXP my_bool xql_comment_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
	DLLEXP void xql_comment_deinit(UDF_INIT *initid);
	DLLEXP char *xql_comment(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);

	DLLEXP my_bool xql_cdata_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
	DLLEXP void xql_cdata_deinit(UDF_INIT *initid);
	DLLEXP char *xql_cdata(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);

	DLLEXP my_bool xql_encode_init(UDF_INIT *initid, UDF_ARGS *args, char *message);
	DLLEXP void xql_encode_deinit(UDF_INIT *initid);
	DLLEXP char *xql_encode(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error);
#ifdef	__cplusplus
}
#endif

#define TRIM_BACKQUOTE(fnull) (fnull+(int)(fnull[0]=='`'))    // Skip starting backquote

/* Memory functions */

/**
 * Dirty hack to prevent MySQL strings from having to be copied to make them 0 terminated.
 * Contributed by Rik Arends.
 */
class cFastNull
{ 
	private:
		char old_char;
		char *char_ptr;
	
	public:
		cFastNull()
		{
			char_ptr = NULL;
		}
	
		cFastNull(char *str, unsigned long i)
		{
			init(str, i);
		}
		
		~cFastNull()
		{
			deinit();
		}
		
		void init(char *str, unsigned long i)
		{
			if (!str) {
				char_ptr = NULL;
				return;
			}
			
			old_char=str[i];
			char_ptr=&str[i];
			str[i]=0;
		}
			
		void deinit()
		{
			if (char_ptr) *char_ptr=old_char;
			char_ptr=0;
		}
};

/**
 * Allocate unused space for an array of nelem elements each of whose size in bytes is elsize.
 * The space shall be initialized to all bits 0.
 */
void **ptr_calloc(size_t nelem, size_t elsize)
{
	void **ptr = (void **)malloc(nelem * elsize + sizeof(int));
	*(int *)ptr = nelem;

	ptr = (void **)((int*)ptr + 1);
	memset(ptr, 0, nelem * elsize);
	
	return ptr;
}

/**
 * Free allocated space of ptr and the space of all items from ptr.
 * Only use with mem allocated with ptr_calloc.
 */
void ptr_free(void **ptr)
{
	for (int i=0; i < *((int *)ptr - 1); i++) {
		if (ptr[i]) free(ptr[i]);
	}
	
	free((int*)ptr-1);
}


/* General functions */

typedef struct
{
	xmlTextWriterPtr writer;
	xmlBufferPtr buffer;
	char add_root;
	char add_child;
	char raw_content;
	char **arg_names;
} xmlWriterBuffer;

xmlWriterBuffer *newXmlWriterBuffer(char *message)
{
	xmlWriterBuffer* data = new xmlWriterBuffer;

	data->buffer = xmlBufferCreate();
	if (data->buffer == NULL) {
		strcpy(message, "Init xql_element failed; Unable to create output buffer");
		return NULL;
	}

	data->writer = xmlNewTextWriterMemory(data->buffer, 0);
	if (data->writer == NULL) {
		strcpy(message, "Init xql_element failed; Unable to create xml text writer");
		xmlBufferFree(data->buffer);
		return NULL;
	}

	data->add_root = 0; 
	
	data->add_child = 0;
	data->raw_content = 0;
	data->arg_names = NULL;

	return data;
}

void freeXmlWriterBuffer(xmlWriterBuffer *data)
{
	xmlBufferFree(data->buffer);
	xmlFreeTextWriter(data->writer);
	if (data->arg_names) ptr_free((void **)data->arg_names);
	
	free(data);
}

char *copyArgName(char *att, int length)
{
	char *attcl = att;
	char *str;
	char quoting = 0;
	
	for (char *ptr=att; ptr<att+length; ptr++) {
		if (*ptr == '`') quoting != quoting;
		 else if (!quoting && *ptr == '.') attcl = ptr+1;
	}
	
	length = length - (attcl-att);
	
	if (!quoting) {
		if (attcl[0] == '`') { attcl++; length--; }
		if (*(attcl+length-1) == '`') length--;
	}
	
	str = (char *)malloc(length + 1);
	if (!str) return NULL;
	
	strncpy(str, attcl, length);
	str[length] = 0;
	
	return str;
}

int strncmp_caseins(char *str1, char *str2, size_t num )
{
	char c1, c2;
	
	for (int i=0; i<num; i++) {
		c1 = (str1[i] >= 65 && str1[i] <= 90) ? str1[i] + 32 : str1[i]; /* Change to lower case */
		c2 = (str2[i] >= 65 && str2[i] <= 90) ? str2[i] + 32 : str2[i]; /* Change to lower case */
		
		if (c1 != c2) return (c1 < c2) * -2 + 1;   /* Could have used q?a:b, but... nerd power */
		if (c1 == 0) return 0;
	}
	
	return 0;
}

int charinstr(char *str, char c, size_t num)
{
	for (int i=0; i<num && str[i]; i++) {
		if (str[i] == c) return i;
	}
	
	return -1;
}

/**
 * Output the library version.
 * lib_mysqludf_xql_info_init()
 */

my_bool lib_mysqludf_xql_info_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	return 0;
}

void lib_mysqludf_xql_info_deinit(UDF_INIT *initid)
{
}

char* lib_mysqludf_xql_info(UDF_INIT *initid, UDF_ARGS *args, char* result, unsigned long* length,	char *is_null, char *error)
{
	strcpy(result, LIBVERSION);
	*length = strlen(LIBVERSION);
	return result;
}

/**
 * A simple xmlwriter test function.
 * xql_test()
 */

my_bool xql_test_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	xmlWriterBuffer* data;

	data = newXmlWriterBuffer(message);
	if (data == NULL) return 1;

	initid->ptr = (char*)data;
	return 0;
}

void xql_test_deinit(UDF_INIT *initid)
{
	freeXmlWriterBuffer((xmlWriterBuffer*)initid->ptr);
}

char *xql_test(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	xmlTextWriterPtr writer;
	xmlBufferPtr buffer;

	writer = ((xmlWriterBuffer*)initid->ptr)->writer;
	buffer = ((xmlWriterBuffer*)initid->ptr)->buffer;
	xmlBufferEmpty(buffer);

	xmlTextWriterStartDocument(writer, "1.0", "ISO-8859-1", NULL);
	xmlTextWriterWriteElement(writer, (xmlChar*)"Hello", (xmlChar*)"World");
	xmlTextWriterEndDocument(writer);
	xmlTextWriterFlush(writer);

	*length = (unsigned long)buffer->use;
	return (char*)buffer->content;
}


/**
 * Write a text declarion; <?xml version="1.0" encoding="iso-8859-1" standalone="yes"?>.
 * xql_textdecl(string version, string encoding [, string standalone])
 */

my_bool xql_textdecl_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	xmlWriterBuffer* data;

	if (args->arg_count != 2 && args->arg_count != 3) {
		strcpy(message,"xql_textdecl() requires at least 2 arguments (version, encoding [, standalone])");
		return 1;
	}

	data = newXmlWriterBuffer(message);
	if (data == NULL) return 1;

	args->arg_type[0] = STRING_RESULT;
	args->arg_type[1] = STRING_RESULT;
	args->arg_type[2] = STRING_RESULT;

	initid->ptr = (char*)data;
	return 0;
}

void xql_textdecl_deinit(UDF_INIT *initid)
{
	freeXmlWriterBuffer((xmlWriterBuffer*)initid->ptr);
}

char *xql_textdecl(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	xmlTextWriterPtr writer;
	xmlBufferPtr buffer;

	if (!args->args[0] || !args->lengths[0] || !args->args[1] || !args->lengths[1]) {
		*error=1;
		return 0;
	}

	writer = ((xmlWriterBuffer*)initid->ptr)->writer;
	buffer = ((xmlWriterBuffer*)initid->ptr)->buffer;
	xmlBufferEmpty(buffer);

	cFastNull *fnull = new cFastNull[args->arg_count];
	for (int i=0; i<args->arg_count; i++) {
		if (args->args[i]) fnull[i].init(args->args[i], args->lengths[i]);
	}

	xmlTextWriterStartDocument(writer, args->args[0], args->args[1], args->arg_count >= 3 && args->args[2] ? args->args[2] : NULL);
	xmlTextWriterFlush(writer);

	delete []fnull;

	*length = (unsigned long)buffer->use;
	return (char*)buffer->content;
}


/**
 * Write a processing instruction.
 * xql_pi(string content AS target)
 */

my_bool xql_pi_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	xmlWriterBuffer* data;

	if (args->arg_count != 1) {
		strcpy(message,"xql_pi() requires 1 argument (content AS target)");
		return 1;
	}

	data = newXmlWriterBuffer(message);
	if (data == NULL) return 1;

	args->arg_type[0] = STRING_RESULT;
	
	initid->ptr = (char*)data;
	return 0;
}

void xql_pi_deinit(UDF_INIT *initid)
{
	freeXmlWriterBuffer((xmlWriterBuffer*)initid->ptr);
}

char *xql_pi(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	xmlTextWriterPtr writer;
	xmlBufferPtr buffer;

	if (!args->args[0]) {
		*error=1;
		return 0;
	}

	writer = ((xmlWriterBuffer*)initid->ptr)->writer;
	buffer = ((xmlWriterBuffer*)initid->ptr)->buffer;
	xmlBufferEmpty(buffer);

	cFastNull fnull_arg(args->args[0], args->lengths[0]);
	cFastNull fnull_att(args->attributes[0], args->attribute_lengths[0]);

	xmlTextWriterWritePI(writer, (xmlChar*)args->attributes[0], (xmlChar*)args->args[0]);
	xmlTextWriterFlush(writer);

	*length = (unsigned long)buffer->use;
	return (char*)buffer->content;
}


/**
 * Write a DTD tag.
 * xql_dtd(string name [, string pubid, string sysid [, string subset]]])
 */

my_bool xql_dtd_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	xmlWriterBuffer* data;

	if (args->arg_count < 1) {
		strcpy(message,"xql_dtd() requires at least 1 argument (name)");
		return 1;
	}

	if (args->arg_count == 2) {
		strcpy(message,"xql_dtd() argument pubid given without argument sysid");
		return 1;
	}

	data = newXmlWriterBuffer(message);
	if (data == NULL) return 1;

	for (int i=0; i<args->arg_count; i++) {
		args->arg_type[i] = STRING_RESULT;
	}

	initid->ptr = (char*)data;
	return 0;
}

void xql_dtd_deinit(UDF_INIT *initid)
{
	freeXmlWriterBuffer((xmlWriterBuffer*)initid->ptr);
}

char *xql_dtd(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	xmlTextWriterPtr writer;
	xmlBufferPtr buffer;

	if (!args->args[0] || (args->arg_count > 1 && (args->args[1] == NULL xor args->args[2] == NULL))) {
		*error=1;
		return 0;
	}

	writer = ((xmlWriterBuffer*)initid->ptr)->writer;
	buffer = ((xmlWriterBuffer*)initid->ptr)->buffer;
	xmlBufferEmpty(buffer);
	
	cFastNull *fnull = new cFastNull[args->arg_count];
	for (int i=0; i<args->arg_count; i++) {
		if (args->args[i]) fnull[i].init(args->args[i], args->lengths[i]);
	}

	xmlTextWriterWriteDTD(writer, (xmlChar*)args->args[0], (xmlChar*)(args->arg_count >= 2 ? args->args[1] : NULL), (xmlChar*)(args->arg_count >= 3 ? args->args[2] : NULL), (xmlChar*)(args->arg_count >= 4 ? args->args[3] : NULL));
	xmlTextWriterFlush(writer);
	
	delete []fnull;
	
	*length = (unsigned long)buffer->use;
	return (char*)buffer->content;
}


/**
 * Write element node with attributes.
 * xql_element(string name, [string content | string xml AS `xql:%` [, string att_val AS `att_name`, ...]])
 */

my_bool xql_element_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	xmlWriterBuffer* data;

	if (args->arg_count < 1) {
		strcpy(message,"xql_element() requires at least 1 argument (name)");
		return 1;
	}

	data = newXmlWriterBuffer(message);
	if (data == NULL) return 1;

	if (args->arg_count > 2) data->arg_names = (char **)ptr_calloc(args->arg_count, sizeof(char *));

	args->arg_type[0] = STRING_RESULT;
	if (args->arg_count > 1) args->arg_type[1] = STRING_RESULT;
	for (int i=2; i<args->arg_count; i++) {
		args->arg_type[i] = STRING_RESULT;
		data->arg_names[i] = copyArgName(args->attributes[i], args->attribute_lengths[i]);
	}

	data->raw_content = strncmp_caseins(TRIM_BACKQUOTE(args->attributes[1]), "xql:", 4) == 0 || strncmp_caseins(args->attributes[1], "xql_", 4) == 0 && charinstr(args->attributes[1], '(', args->attribute_lengths[1]);

	initid->ptr = (char*)data;
	return 0;
}

void xql_element_deinit(UDF_INIT *initid)
{
	freeXmlWriterBuffer((xmlWriterBuffer*)initid->ptr);
}

char *xql_element(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	xmlTextWriterPtr writer;
	xmlBufferPtr buffer;

	if (!args->args[0] || !args->lengths[0]) {
		*is_null=1;
		return 0;
	}

	writer = ((xmlWriterBuffer*)initid->ptr)->writer;
	buffer = ((xmlWriterBuffer*)initid->ptr)->buffer;
	xmlBufferEmpty(buffer);

	cFastNull *fnull = new cFastNull[args->arg_count];
	for (int i=0; i<args->arg_count; i++) {
		if (args->args[i]) fnull[i].init(args->args[i], args->lengths[i]);
	}

	xmlTextWriterStartElement(writer, (xmlChar*)args->args[0]);
	for (int i=2; i<args->arg_count; i++) {
		if (args->args[i]) xmlTextWriterWriteAttribute(writer, (xmlChar*)TRIM_BACKQUOTE(args->attributes[i]), (xmlChar*)args->args[i]);
	}
	
	if (args->arg_count > 1 && args->args[1]) {
		if (((xmlWriterBuffer*)initid->ptr)->raw_content) {
			xmlTextWriterWriteRaw(writer, (xmlChar*)args->args[1]);
		} else {
			xmlTextWriterWriteString(writer, (xmlChar*)args->args[1]);
		}
	}
	xmlTextWriterEndElement(writer);
	xmlTextWriterFlush(writer);

	delete []fnull;

	*length = (unsigned long)buffer->use;
	return (char*)buffer->content;
}


/**
 * Write an element for each name/content pair.
 * Arguments starting with `xql:` as name or output of xql functions are inserted as raw XML.
 * 
 * xql_forest([string name_root AS `xql:root`,] string content AS name | string xml, ...)
 */

my_bool xql_forest_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	xmlWriterBuffer* data;

	data = newXmlWriterBuffer(message);
	if (data == NULL) return 1;

	data->arg_names = (char **)ptr_calloc(args->arg_count, sizeof(char *));

	for (int i=0; i<args->arg_count; i++) {
		args->arg_type[i] = STRING_RESULT;
		data->arg_names[i] = copyArgName(args->attributes[i], args->attribute_lengths[i]);
	}
	
	data->add_root = strncmp_caseins(args->attributes[0], "xql:root", args->attribute_lengths[0]) == 0;

	initid->ptr = (char*)data;
	return 0;
}

void xql_forest_deinit(UDF_INIT *initid)
{
	freeXmlWriterBuffer((xmlWriterBuffer*)initid->ptr);
}

char *xql_forest(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	xmlTextWriterPtr writer;
	xmlBufferPtr buffer;
	char has_root=0;

	if (args->arg_count == 0) {
		*is_null = 1;
		return 0;
	}

	writer = ((xmlWriterBuffer*)initid->ptr)->writer;
	buffer = ((xmlWriterBuffer*)initid->ptr)->buffer;
	xmlBufferEmpty(buffer);

	cFastNull *fnull = new cFastNull[args->arg_count];
	for (int i=0; i<args->arg_count; i++) {
		if (args->args[i]) fnull[i].init(args->args[i], args->lengths[i]);
	}

	if (((xmlWriterBuffer*)initid->ptr)->add_root) xmlTextWriterStartElement(writer, (xmlChar*)args->args[0]);

	for (int i=(int)((xmlWriterBuffer*)initid->ptr)->add_root; i<args->arg_count; i++) {
		if (args->args[i]) {
			if (!((xmlWriterBuffer*)initid->ptr)->arg_names[i]) {
				xmlTextWriterWriteRaw(writer, (xmlChar*)args->args[i]);
			} else {
				xmlTextWriterWriteElement(writer, (xmlChar*)((xmlWriterBuffer*)initid->ptr)->arg_names[i], (xmlChar*)args->args[i]);
			}
		}
	}

	if (((xmlWriterBuffer*)initid->ptr)->add_root) xmlTextWriterEndElement(writer);
	xmlTextWriterFlush(writer);

	delete []fnull;

	*length = (unsigned long)buffer->use;
	return (char*)buffer->content;
}


/**
 * Concatenate xml.
 * xql_concat([string name_root AS `xql:node`,] string xml, ...)
 */

my_bool xql_concat_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	xmlWriterBuffer* data;

	data = newXmlWriterBuffer(message);
	if (data == NULL) return 1;

	for (int i=0; i<args->arg_count; i++) {
		args->arg_type[i] = STRING_RESULT;
	}

	data->add_root = strncmp_caseins(args->attributes[0], "xql:root", args->attribute_lengths[0]) == 0;

	initid->ptr = (char*)data;
	return 0;
}

void xql_concat_deinit(UDF_INIT *initid)
{
	freeXmlWriterBuffer((xmlWriterBuffer*)initid->ptr);
}

char *xql_concat(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	xmlTextWriterPtr writer;
	xmlBufferPtr buffer;

	writer = ((xmlWriterBuffer*)initid->ptr)->writer;
	buffer = ((xmlWriterBuffer*)initid->ptr)->buffer;
	xmlBufferEmpty(buffer);

	*is_null = !((xmlWriterBuffer*)initid->ptr)->add_root;

	cFastNull *fnull = new cFastNull[args->arg_count];
	for (int i=0; i<args->arg_count; i++) {
		if (args->args[i]) fnull[i].init(args->args[i], args->lengths[i]);
	}

	if (((xmlWriterBuffer*)initid->ptr)->add_root) xmlTextWriterStartElement(writer, (xmlChar*)args->args[0]);

	for (int i=0; i<args->arg_count; i++) {
		if (args->args[i]) {
			args->args[i][args->lengths[i]] = 0;
			xmlTextWriterWriteRaw(writer, (xmlChar*)args->args[i]);
			*is_null = 0;
		}
	}

	if (((xmlWriterBuffer*)initid->ptr)->add_root) xmlTextWriterEndElement(writer);

	if (*is_null) return 0;

	xmlTextWriterFlush(writer);
	
	delete []fnull;

	*length = (unsigned long)buffer->use;
	return (char*)buffer->content;
}


/**
 * Group concatenate XML for each row.
 * Normal arguments are inserted as XML nodes. Arguments starting with `xql:` as name or output of xql functions are inserted as raw XML.
 * 
 * xql_agg([string name_root AS `xql:root`,] [string name_child AS `xql:child`,] string content AS name | string xml, ...)
 */
my_bool xql_agg_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	xmlWriterBuffer* data;

	data = newXmlWriterBuffer(message);
	if (data == NULL) return 1;

	data->arg_names = (char **)ptr_calloc(args->arg_count, sizeof(char *));

	for (int i=0; i<args->arg_count; i++) {
		args->arg_type[i] = STRING_RESULT;
		
		if (!(strncmp_caseins(TRIM_BACKQUOTE(args->attributes[i]), "xql:", 4) == 0 || strncmp_caseins(args->attributes[i], "xql_", 4) == 0 && charinstr(args->attributes[i], '(', args->attribute_lengths[i]))) {
			data->arg_names[i] = copyArgName(args->attributes[i], args->attribute_lengths[i]);
			if (!data->arg_names[i]) {
				strcpy(message, "Could not allocate memory to save argument name.");
				return 1;
			}
		}
	}

	data->add_root = strncmp_caseins(args->attributes[0], "xql:root", args->attribute_lengths[0]) == 0;
	data->add_child = strncmp_caseins(args->attributes[data->add_root], "xql:child", args->attribute_lengths[data->add_root]) == 0;	
	
	initid->ptr = (char*)data;
	return 0;
}

void xql_agg_deinit(UDF_INIT *initid)
{
	freeXmlWriterBuffer((xmlWriterBuffer*)initid->ptr);
}

/* This is only for MySQL 4.0 compability */
void axql_agg_reset(UDF_INIT* initid, UDF_ARGS* args, char* is_null, char* message)
{
	xql_agg_clear(initid, is_null, message);
	xql_agg_add(initid, args, is_null, message);
}

void xql_agg_clear(UDF_INIT* initid, char* is_null, char* message)
{
	xmlBufferEmpty(((xmlWriterBuffer*)initid->ptr)->buffer);
	if (((xmlWriterBuffer*)initid->ptr)->add_root != 0) ((xmlWriterBuffer*)initid->ptr)->add_root = 1;
}

void xql_agg_add(UDF_INIT *initid, UDF_ARGS *args, char *is_null, char *error)
{
	xmlTextWriterPtr writer;
	int i = 0;
	char add_is_null = 1;
	char added_child = 0;

	for (i=((xmlWriterBuffer*)initid->ptr)->add_root >= 0 ? 0 : 1; i<args->arg_count; i++) {
		if (args->args[i]) add_is_null = 0;
	}
	if (add_is_null) return;
	
	cFastNull *fnull = new cFastNull[args->arg_count];
	for (int i=0; i<args->arg_count; i++) {
		if (args->args[i]) fnull[i].init(args->args[i], args->lengths[i]);
	}

	i = 0;
	writer = ((xmlWriterBuffer*)initid->ptr)->writer;

	if (((xmlWriterBuffer*)initid->ptr)->add_root == 1) {
		if (args->args[i]) {
			xmlTextWriterStartElement(writer, (xmlChar*)args->args[i]);
			((xmlWriterBuffer*)initid->ptr)->add_root = -2;
		} else {
			((xmlWriterBuffer*)initid->ptr)->add_root = -1;
		}
	}
	
	if (((xmlWriterBuffer*)initid->ptr)->add_root != 0) i++;
	
	if (((xmlWriterBuffer*)initid->ptr)->add_child) {
		if (args->args[i]) {
			added_child = 1;
			xmlTextWriterStartElement(writer, (xmlChar*)args->args[i]);
		}
		i++;		
	}

	for (; i<args->arg_count; i++) {
		if (args->args[i]) {
			if (!((xmlWriterBuffer*)initid->ptr)->arg_names[i]) {
				xmlTextWriterWriteRaw(writer, (xmlChar*)args->args[i]);
			} else {
				xmlTextWriterWriteElement(writer, (xmlChar*)((xmlWriterBuffer*)initid->ptr)->arg_names[i], (xmlChar*)args->args[i]);
			}
		}
	}

	if (added_child) xmlTextWriterEndElement(writer);
	
	delete []fnull;
}

char *xql_agg(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	xmlTextWriterPtr writer;
	xmlBufferPtr buffer;
	
	writer = ((xmlWriterBuffer*)initid->ptr)->writer;
	buffer = ((xmlWriterBuffer*)initid->ptr)->buffer;

	xmlTextWriterFlush(writer);

	// Aggregated empty strings should return an empty string, but now return NULL instead
	if ((unsigned long)((xmlWriterBuffer*)initid->ptr)->buffer->use == 0) {
		*is_null = 1;
		return 0;
	}

	if (((xmlWriterBuffer*)initid->ptr)->add_root == -2) {
		xmlTextWriterEndElement(writer);
		xmlTextWriterFlush(writer);
	}

	*length = (unsigned long)buffer->use;
	return (char*)buffer->content;
}

/**
 * Write comment node.
 * xql_comment(string content)
 */

my_bool xql_comment_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	xmlWriterBuffer* data;

	if (args->arg_count != 1) {
		strcpy(message,"xql_comment() requires 1 argument");
		return 1;
	}

	data = newXmlWriterBuffer(message);
	if (data == NULL) return 1;

	args->arg_type[0] = STRING_RESULT;

	initid->ptr = (char*)data;
	return 0;
}

void xql_comment_deinit(UDF_INIT *initid)
{
	freeXmlWriterBuffer((xmlWriterBuffer*)initid->ptr);
}

char *xql_comment(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	xmlTextWriterPtr writer;
	xmlBufferPtr buffer;

	if (!args->args[0]) {
		*is_null=1;
		return 0;
	}

	writer = ((xmlWriterBuffer*)initid->ptr)->writer;
	buffer = ((xmlWriterBuffer*)initid->ptr)->buffer;
	xmlBufferEmpty(buffer);

	cFastNull fnull_arg(args->args[0], args->lengths[0]);

	xmlTextWriterWriteComment(writer, (xmlChar*)args->args[0]);
	xmlTextWriterFlush(writer);

	*length = (unsigned long)buffer->use;
	return (char*)buffer->content;
}


/**
 * Write CDATA section.
 * xql_cdata(string content)
 */

my_bool xql_cdata_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	xmlWriterBuffer* data;

	if (args->arg_count != 1) {
		strcpy(message,"xql_cdata() requires 1 argument");
		return 1;
	}

	data = newXmlWriterBuffer(message);
	if (data == NULL) return 1;

	args->arg_type[0] = STRING_RESULT;

	initid->ptr = (char*)data;
	return 0;
}

void xql_cdata_deinit(UDF_INIT *initid)
{
	freeXmlWriterBuffer((xmlWriterBuffer*)initid->ptr);
}

char *xql_cdata(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	xmlTextWriterPtr writer;
	xmlBufferPtr buffer;

	if (!args->args[0]) {
		*is_null=1;
		return 0;
	}

	writer = ((xmlWriterBuffer*)initid->ptr)->writer;
	buffer = ((xmlWriterBuffer*)initid->ptr)->buffer;
	xmlBufferEmpty(buffer);

	cFastNull fnull_arg(args->args[0], args->lengths[0]);

	xmlTextWriterWriteCDATA(writer, (xmlChar*)args->args[0]);
	xmlTextWriterFlush(writer);

	*length = (unsigned long)buffer->use;
	return (char*)buffer->content;
}


/**
 * Encode special chars
 * xql_encode(string content)
 */

my_bool xql_encode_init(UDF_INIT *initid, UDF_ARGS *args, char *message)
{
	if (args->arg_count != 1) {
		strcpy(message,"xql_encode() requires 1 argument");
		return 1;
	}

	args->arg_type[0] = STRING_RESULT;
	return 0;
}

void xql_encode_deinit(UDF_INIT *initid)
{
	if (initid->ptr) free(initid->ptr);
}

char *xql_encode(UDF_INIT *initid, UDF_ARGS *args, char *result, unsigned long *length, char *is_null, char *error)
{
	xmlTextWriterPtr writer;
	xmlBufferPtr buffer;

	if (!args->args[0]) {
		*is_null=1;
		return 0;
	}

	if (initid->ptr) free(initid->ptr);

	cFastNull fnull_arg(args->args[0], args->lengths[0]);

	initid->ptr = (char*)xmlEncodeSpecialChars(NULL, (xmlChar*)args->args[0]);
	if (!initid->ptr) {
		*error=1;
		return 0;
	}

	*length = strnlen(initid->ptr, (unsigned long)-1);
	return initid->ptr;
}

#endif /* HAVE_DLOPEN */
